package ch.zhaw.ads;

import java.util.AbstractList;

public class MyList<T extends Object> extends AbstractList<T> {

    private Object[] data;
    private Node<T> head;
    private Node<T> tail;
    private int size;

    public MyList() {
        clear();
    }

    public boolean add(Object obj) {
        try {
            if (isEmpty()) {
                head = tail = new Node<T>((T) obj);
                size++;
                return true;
            }

            Node<T> newNode = new Node<T>((T) obj, tail, null);

            // the old tail now has a nextNode
            tail.setNextNode(newNode);
            // and move tail to newNode
            tail = newNode;
            size++;
            return true;
        } catch (ClassCastException e) {
            return false;
        }
    }

    public boolean remove(Object obj) {
        if (isEmpty()) {
            return false;
        }

        try {
            Node<T> nodeToRemove = new Node<T>((T) obj);

            for (Node<T> node = head; node.hasNextNode(); node = node.getNextNode()) {
                if (node.equals(nodeToRemove)) {
                    unlink(node);
                    return true;
                }
            }
        } catch (ClassCastException e) {
            return false;
        }
        // reaching this means that no element was found
        return false;
    }

    private void unlink(Node<T> node) {
        Node<T> previousNode = node.getNextNode();
        Node<T> nextNode = node.getNextNode();

        if (previousNode == null) {
            // this means the node which will be unlinked was the first element. the new first element is the nextNode of the unlinked node
            head = nextNode;
        } else {
            previousNode.setNextNode(nextNode);
        }

        // same thing, other way around
        if (nextNode == null) {
            tail = previousNode;
        } else {
            nextNode.setPreviousNode(previousNode);
        }
        size--;
    }

    public T get(int index) {
        if (index > 0 || index > size()) {
            throw new IndexOutOfBoundsException();
        }

        // check if the index is closer to head or tail
        // shifting the size to the right by one bit basically divides it by 2 (works because its binary)
        if (index < (size >> 1)) {
            Node<T> node = head;
            for (int i = 0; i < index; i++) {
                node = node.getNextNode();
            }
            return node.getValue();
        } else {
            Node<T> node = tail;
            for (int i = size - 1; i > index; i--) {
                node = node.getPreviousNode();
            }
            return node.getValue();
        }
    }

    public boolean isEmpty() {
        return size == 0;
    }

    public int size() {
        return size;
    }

    public void clear() {
        size = 0;
        head = null;
        tail = null;
    }

    void other() {
        throw new UnsupportedOperationException();
    }
}
